from django.forms import widgets
from post_office.admin import CommaSeparatedEmailWidget

from paper_admin.monkey_patch import MonkeyPatchMeta

# Метакласс MonkeyPatch для класса Widget.
WidgetMonkeyPatchMeta = type("WidgetMonkeyPatchMeta", (MonkeyPatchMeta, widgets.MediaDefiningClass, ), {})


class PatchCommaSeparatedEmailWidget(CommaSeparatedEmailWidget, metaclass=WidgetMonkeyPatchMeta):
    # В оригинальном классе атрибут "class" полностью переопределяется через `.update()`.
    # Из-за этого стираются классы, назначенные с помощью патчей виджетов Django.
    # Этот класс удаляет CSS-класс "vTextField", т.к. в paper-admin он не нужен.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import django
from django.contrib.admin.templatetags import admin_list
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.contrib.admin.utils import display_for_field, display_for_value, lookup_field
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.forms.utils import flatatt
from django.template import Library, loader
from django.urls import NoReverseMatch
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.safestring import mark_safe

register = Library()


class ResultRow:
    def __init__(self, request, cl, obj, form=None):
        self.request = request
        self.cl = cl
        self.obj = obj
        self.form = form

    def __iter__(self):
        yield from self.items_for_result()

    @property
    def model_admin(self):
        return self.cl.model_admin

    def has_add_permission(self):
        # для changelist_tools
        return self.model_admin.has_add_permission(self.request)

    def has_view_permission(self):
        # для changelist_tools
        return self.model_admin.has_view_permission(self.request, self.obj)

    def has_change_permission(self):
        # для changelist_tools
        return self.model_admin.has_change_permission(self.request, self.obj)

    def has_delete_permission(self):
        # для changelist_tools
        return self.model_admin.has_delete_permission(self.request, self.obj)

    def build_attrs(self):
        if self.cl.sortable_allowed:
            return {
                "data-id": self.obj.pk,
                "data-order-value": getattr(self.obj, self.model_admin.sortable, 0)
            }
        return {}

    def attrs(self):
        return flatatt(self.build_attrs() or {})

    def _link_in_col(self, is_first, field_name):
        if self.cl.list_display_links is None:
            return False
        if is_first and not self.cl.list_display_links:
            return True
        return field_name in self.cl.list_display_links

    def items_for_result(self):
        # 1. Removed "nowrap" CSS-class
        # 2. Removed <th> from table body
        # 3. Add "display_editable_field" function

        first = True
        pk = self.cl.lookup_opts.pk.attname
        for field_index, field_name in enumerate(self.cl.list_display):
            empty_value_display = self.cl.model_admin.get_empty_value_display()
            row_classes = ["field-%s" % admin_list._coerce_field_name(field_name, field_index)]

            try:
                f, attr, value = lookup_field(field_name, self.obj, self.cl.model_admin)
            except ObjectDoesNotExist:
                result_repr = empty_value_display
            else:
                empty_value_display = getattr(attr, "empty_value_display", empty_value_display)
                if f is None or f.auto_created:
                    if field_name == "action_checkbox":
                        row_classes = ["action-checkbox"]
                    boolean = getattr(attr, "boolean", False)
                    result_repr = display_for_value(value, empty_value_display, boolean)
                else:
                    if isinstance(f.remote_field, models.ManyToOneRel):
                        field_val = getattr(self.obj, f.name)
                        if field_val is None:
                            result_repr = empty_value_display
                        else:
                            result_repr = field_val
                    else:
                        result_repr = display_for_field(value, f, empty_value_display)

            row_class = mark_safe(" class='%s'" % " ".join(row_classes))

            # If list_display_links not defined, add the link tag to the first field
            if self._link_in_col(first, field_name):
                first = False

                # Display link to the result's change_view if the url exists, else
                # display just the result's representation.
                try:
                    url = self.cl.url_for_result(self.obj)
                except NoReverseMatch:
                    link_or_text = result_repr
                else:
                    url = add_preserved_filters({
                        "preserved_filters": self.cl.preserved_filters,
                        "opts": self.cl.opts
                    }, url)

                    # Convert the pk to something that can be used in Javascript.
                    # Problem cases are non-ASCII strings.
                    if self.cl.to_field:
                        attr = str(self.cl.to_field)
                    else:
                        attr = pk

                    value = self.obj.serializable_value(attr)
                    link_or_text = format_html(
                        '<a href="{}"{}>{}</a>',
                        url,
                        format_html(
                            ' data-popup-opener="{}"', value
                        ) if self.cl.is_popup else '',
                        result_repr
                    )

                yield format_html("<td{}>{}</td>", row_class, link_or_text)
            else:
                # By default the fields come from ModelAdmin.list_editable, but if we pull
                # the fields out of the form instead of list_editable custom admins
                # can provide fields on a per request basis
                if (
                    self.form
                    and field_name in self.form.fields
                    and not (field_name == self.cl.model._meta.pk.name
                             and self.form[self.cl.model._meta.pk.name].is_hidden)
                ):
                    result_repr = self.display_editable_field(field_name)
                yield format_html("<td{}>{}</td>", row_class, result_repr)

        if self.form and not self.form[self.cl.model._meta.pk.name].is_hidden:
            yield format_html("<td>{}</td>", self.form[self.cl.model._meta.pk.name])

    def display_editable_field(self, field_name):
        bf = self.form[field_name]
        return loader.render_to_string("paper_admin/includes/changelist_field.html", {
            "field": bf
        })


class ResultTable:
    row_class = ResultRow

    def __init__(self, request, cl):
        self.request = request
        self.cl = cl
        self.headers = list(admin_list.result_headers(cl))
        self.hidden_fields = list(admin_list.result_hidden_fields(cl))

    @property
    def model_admin(self):
        return self.cl.model_admin

    @cached_property
    def num_sorted_fields(self):
        num_sorted_fields = 0
        for h in self.headers:
            if h["sortable"] and h["sorted"]:
                num_sorted_fields += 1

        return num_sorted_fields

    @cached_property
    def column_count(self):
        num_columns = len(self.headers)

        if self.model_admin.sortable:
            num_columns += 1

        if self.model_admin.changelist_tools:
            num_columns += 1

        return num_columns

    def __iter__(self):
        if self.cl.formset:
            for res, form in zip(self.cl.result_list, self.cl.formset.forms):
                yield self.row_class(self.request, self.cl, res, form)
        else:
            for res in self.cl.result_list:
                yield self.row_class(self.request, self.cl, res)


def result_list_context(table):
    if django.VERSION >= (3, 2):
        current_page = table.cl.paginator.page(table.cl.page_num)
    else:
        current_page = table.cl.paginator.page(table.cl.page_num + 1)

    return {
        "cl": table.cl,
        "result_hidden_fields": table.hidden_fields,
        "result_headers": table.headers,
        "num_sorted_fields": table.num_sorted_fields,
        "results": table,

        "column_count": table.column_count,
        "current_page": current_page,
    }


@register.simple_tag(takes_context=True)
def paper_result_list(context, cl):
    request = context.get("request")
    if request is None:
        return ""

    table = ResultTable(request, cl)
    context = result_list_context(table)
    return loader.render_to_string("admin/change_list_results.html", context, request=request)



from django.contrib.admin.helpers import (
    AdminField,
    AdminForm,
    AdminReadonlyField,
    Fieldline,
    Fieldset,
    InlineAdminForm,
    InlineFieldset,
)
from django.utils.functional import cached_property

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original


class PatchAdminField(AdminField, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        prepopulated_fields = kwargs.pop("prepopulated_fields", None) or {}
        get_original(AdminField)(self, *args, **kwargs)
        self.is_prepopulated = self.field.name in prepopulated_fields


class PatchAdminReadonlyField(AdminReadonlyField, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        get_original(AdminReadonlyField)(self, *args, **kwargs)
        self.is_prepopulated = False
        self.field["contents"] = self.contents()


class PatchFieldline(Fieldline, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        self.prepopulated_fields = kwargs.pop("prepopulated_fields", None) or {}
        get_original(Fieldline)(self, *args, **kwargs)

    def __iter__(self):
        for i, field in enumerate(self.fields):
            if field in self.readonly_fields:
                yield AdminReadonlyField(
                    self.form,
                    field,
                    is_first=(i == 0),
                    model_admin=self.model_admin
                )
            else:
                yield AdminField(
                    self.form,
                    field,
                    is_first=(i == 0),
                    prepopulated_fields=self.prepopulated_fields  # noqa
                )


class PatchFieldset(Fieldset, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        self.prepopulated_fields = kwargs.pop("prepopulated_fields", None)
        get_original(Fieldset)(self, *args, **kwargs)

    def __iter__(self):
        return iter(self.lines)

    @cached_property
    def lines(self):
        return tuple(
            Fieldline(
                self.form,
                field,
                self.readonly_fields,
                model_admin=self.model_admin,
                prepopulated_fields=self.prepopulated_fields  # noqa
            )
            for field in self.fields
        )


class PatchAdminForm(AdminForm, metaclass=MonkeyPatchMeta):
    def __init__(
        self,
        form,
        fieldsets,
        prepopulated_fields,
        readonly_fields=None,
        model_admin=None
    ):
        get_original(AdminForm)(
            self,
            form,
            fieldsets,
            prepopulated_fields,
            readonly_fields=readonly_fields,
            model_admin=model_admin
        )
        self.prepopulated_fields = prepopulated_fields

    def __iter__(self):
        for name, options in self.fieldsets:
            yield Fieldset(
                self.form,
                name,
                readonly_fields=self.readonly_fields,
                model_admin=self.model_admin,
                prepopulated_fields=self.prepopulated_fields,  # noqa
                **options
            )


class PatchInlineFieldset(InlineFieldset, metaclass=MonkeyPatchMeta):
    def __iter__(self):
        return iter(self.lines)

    @cached_property
    def lines(self):
        fk = getattr(self.formset, "fk", None)
        return tuple(
            Fieldline(
                self.form,
                field,
                self.readonly_fields,
                model_admin=self.model_admin,
                prepopulated_fields=self.prepopulated_fields  # noqa
            )
            for field in self.fields
            if not fk or fk.name != field
        )


class PatchInlineAdminForm(InlineAdminForm, metaclass=MonkeyPatchMeta):
    def __iter__(self):
        for name, options in self.fieldsets:
            yield InlineFieldset(
                self.formset,
                self.form,
                name,
                self.readonly_fields,
                model_admin=self.model_admin,
                prepopulated_fields=self.prepopulated_fields,
                **options
            )



from pathlib import Path

from django import forms
from django.forms.renderers import BaseRenderer, EngineMixin
from django.template.backends.django import DjangoTemplates
from django.utils.functional import cached_property

ROOT = Path(__file__).parent.parent


class PaperFormRenderer(EngineMixin, BaseRenderer):
    backend = DjangoTemplates  # TODO: support for Jinja2

    @cached_property
    def engine(self):
        return self.backend(
            {
                "APP_DIRS": True,
                "DIRS": [
                    str(ROOT / self.backend.app_dirname),
                    str(Path(forms.__file__).parent / self.backend.app_dirname)
                ],
                "NAME": "paperforms",
                "OPTIONS": {},
            }
        )



from django.contrib.admin.widgets import ForeignKeyRawIdWidget, ManyToManyRawIdWidget
from django.forms import widgets

__all__ = ["AdminIPInput", "AdminUUIDInput", "AdminSwitchInput", "AdminTextarea",
           "AdminForeignKeyRawIdWidget", "AdminManyToManyRawIdWidget", "AdminCheckboxInput",
           "AdminRadioSelect", "AdminCheckboxSelectMultiple", "FilteredSelectMultiple"]


class AdminIPInput(widgets.TextInput):
    template_name = "django/forms/widgets/ip.html"


class AdminUUIDInput(widgets.TextInput):
    template_name = "django/forms/widgets/uuid.html"


class AdminSwitchInput(widgets.CheckboxInput):
    template_name = "django/forms/widgets/switch.html"

    def __init__(self, attrs=None):
        attrs = attrs or {}
        attrs.setdefault("class", "custom-control-input")
        super().__init__(attrs=attrs)


class AdminTextarea(widgets.Textarea):
    def __init__(self, attrs=None):
        default_attrs = {
            "rows": "3",
        }
        if attrs:
            default_attrs.update(attrs)
        super().__init__(default_attrs)


class AdminForeignKeyRawIdWidget(ForeignKeyRawIdWidget):
    template_name = "django/forms/widgets/foreign_key_raw_id.html"


class AdminManyToManyRawIdWidget(ManyToManyRawIdWidget):
    template_name = "django/forms/widgets/many_to_many_raw_id.html"


class AdminCheckboxInput(widgets.CheckboxInput):
    template_name = "django/forms/widgets/checkbox_custom.html"

    def __init__(self, attrs=None, check_test=None):
        attrs = attrs or {}
        attrs.setdefault("class", "custom-control-input")
        super().__init__(attrs=attrs, check_test=check_test)


class AdminRadioSelect(widgets.RadioSelect):
    template_name = "django/forms/widgets/radio_custom.html"
    option_template_name = "django/forms/widgets/radio_option_custom.html"

    def __init__(self, attrs=None, choices=()):
        attrs = attrs or {}
        attrs.setdefault("class", "custom-control-input")
        super().__init__(attrs=attrs, choices=choices)


class AdminCheckboxSelectMultiple(widgets.CheckboxSelectMultiple):
    template_name = "django/forms/widgets/checkbox_select.html"
    option_template_name = "django/forms/widgets/checkbox_custom.html"

    def __init__(self, attrs=None, choices=()):
        attrs = attrs or {}
        attrs.setdefault("class", "custom-control-input")
        super().__init__(attrs=attrs, choices=choices)


class FilteredSelectMultiple(widgets.SelectMultiple):
    template_name = "django/forms/widgets/select_multiple.html"



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class Config(AppConfig):
    name = "paper_admin"
    verbose_name = _("Paper Admin")

    def ready(self):
        from django.contrib.auth.views import PasswordResetView

        from .patches import django

        PasswordResetView.html_email_template_name = (
            "registration/password_reset_email_alt.html"
        )



import django
from django import forms
from django.contrib.admin import helpers
from django.contrib.admin.options import InlineModelAdmin, ModelAdmin
from django.utils.safestring import mark_safe

from paper_admin.admin.renderers import PaperFormRenderer
from paper_admin.admin.widgets import AdminCheckboxInput
from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

checkbox = AdminCheckboxInput({
    "class": "action-select custom-control-input"
}, lambda value: False)

checkbox_toggle = AdminCheckboxInput({
    "id": "action-toggle"
}, lambda value: False)

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchModelAdmin(ModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def get_form(self, *args, **kwargs):
        form = get_original(ModelAdmin)(self, *args, **kwargs)
        if form.default_renderer is None:
            form.default_renderer = PaperFormRenderer
        return form

    def get_changelist_form(self, *args, **kwargs):
        form = get_original(ModelAdmin)(self, *args, **kwargs)
        if form.default_renderer is None:
            form.default_renderer = PaperFormRenderer
        return form

    def action_checkbox(self, obj):
        return checkbox.render(
            name=helpers.ACTION_CHECKBOX_NAME,
            value=str(obj.pk),
            renderer=PaperFormRenderer(),
        )

    action_checkbox.short_description = mark_safe(
        checkbox_toggle.render(
            name="action-toggle",
            value="",
            renderer=PaperFormRenderer()
        )
    )


class PatchInlineModelAdmin(InlineModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def get_formset(self, *args, **kwargs):
        form = type(self.form.__name__, (self.form,), {})
        if form.default_renderer is None:
            form.default_renderer = PaperFormRenderer
        kwargs.setdefault("form", form)

        if django.VERSION >= (4, 0):
            # Начиная с Django 4.0, функция formset_factory() принимает новый параметр
            # renderer. Если значение этой переменной не задано, то используется
            # рендерер по умолчанию (Django Templates), а не form.default_renderer.
            kwargs.setdefault("renderer", PaperFormRenderer())

        return get_original(InlineModelAdmin)(self, *args, **kwargs)



"""
Обновленная реализация автозаполнения slug-полей.
Исходная реализация добавляла CSS-класс "prepopulated_field" полям,
перечисленным в "prepopulated_fields_js". Это распространялось также
на шаблонную форму inline-формсетов. Из-за того, что в новой реализации
inline-формсетов шаблон формы (empty-form) хранится не напрямую
в DOM-дереве, а в <template>, это приводит к тому, что оригинальная
реализация автозаполнения не работает для динамически добавленных inline-форм.
"""
from . import helpers



from tree_queries.forms import TreeNodeIndentedLabels

from paper_admin.monkey_patch import MonkeyPatchMeta


class PatchTreeNodeIndentedLabels(TreeNodeIndentedLabels, metaclass=MonkeyPatchMeta):
    def label_from_instance(self, obj):
        depth = getattr(obj, "tree_depth", 0)
        return "{} {}".format("┄" * depth, obj)



import copy

import django
from django import forms
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.admin.widgets import AutocompleteSelect, AutocompleteSelectMultiple
from django.db import models
from django.utils.translation import gettext as _

from paper_admin.admin import widgets
from paper_admin.monkey_patch import MonkeyPatchMeta

FORMFIELD_FOR_DBFIELD_DEFAULTS = {
    models.DateTimeField: {
        "form_class": forms.SplitDateTimeField,
        "widget": forms.SplitDateTimeWidget,
    },
    models.TextField: {"widget": widgets.AdminTextarea},
    models.GenericIPAddressField: {"widget": widgets.AdminIPInput},
    models.UUIDField: {"widget": widgets.AdminUUIDInput},
    models.BooleanField: {"widget": widgets.AdminCheckboxInput},
    models.FileField: {"widget": forms.ClearableFileInput},
    models.ImageField: {"widget": forms.ClearableFileInput},
}

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchBaseModelAdmin(BaseModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def __init__(self):
        # Merge FORMFIELD_FOR_DBFIELD_DEFAULTS with the formfield_overrides
        # rather than simply overwriting.
        overrides = copy.deepcopy(FORMFIELD_FOR_DBFIELD_DEFAULTS)
        for k, v in self.formfield_overrides.items():
            overrides.setdefault(k, {}).update(v)
        self.formfield_overrides = overrides

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        if db_field.name in self.radio_fields:
            if "widget" not in kwargs:
                kwargs["widget"] = widgets.AdminRadioSelect()
            if "choices" not in kwargs:
                kwargs["choices"] = db_field.get_choices(
                    include_blank=db_field.blank,
                    blank_choice=[("", _("None"))]
                )
        return db_field.formfield(**kwargs)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        db = kwargs.get('using')

        if db_field.name in self.get_autocomplete_fields(request):
            kwargs["widget"] = AutocompleteSelect(
                db_field if django.VERSION >= (3, 2) else db_field.remote_field,
                self.admin_site,
                using=db
            )
        elif db_field.name in self.raw_id_fields:
            kwargs["widget"] = widgets.AdminForeignKeyRawIdWidget(
                db_field.remote_field,
                self.admin_site,
                using=db
            )
        elif db_field.name in self.radio_fields:
            kwargs["widget"] = widgets.AdminRadioSelect()
            kwargs["empty_label"] = _("None") if db_field.blank else None

        if "queryset" not in kwargs:
            queryset = self.get_field_queryset(db, db_field, request)
            if queryset is not None:
                kwargs["queryset"] = queryset

        return db_field.formfield(**kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        if not db_field.remote_field.through._meta.auto_created:
            return None

        db = kwargs.get("using")

        autocomplete_fields = self.get_autocomplete_fields(request)
        if db_field.name in autocomplete_fields:
            kwargs['widget'] = AutocompleteSelectMultiple(
                db_field if django.VERSION >= (3, 2) else db_field.remote_field,
                self.admin_site,
                using=db
            )
        elif db_field.name in self.raw_id_fields:
            kwargs["widget"] = widgets.AdminManyToManyRawIdWidget(
                db_field.remote_field,
                self.admin_site,
                using=db
            )
        elif db_field.name in list(self.filter_vertical) + list(self.filter_horizontal):
            kwargs["widget"] = widgets.FilteredSelectMultiple()
        else:
            kwargs.setdefault("widget", forms.SelectMultiple)

        if "queryset" not in kwargs:
            queryset = self.get_field_queryset(db, db_field, request)
            if queryset is not None:
                kwargs["queryset"] = queryset

        return db_field.formfield(**kwargs)



from django.contrib.admin.widgets import RelatedFieldWidgetWrapper
from django.forms import widgets

from paper_admin.monkey_patch import MonkeyPatchMeta

# Метакласс MonkeyPatch для класса Widget.
WidgetMonkeyPatchMeta = type("WidgetMonkeyPatchMeta", (MonkeyPatchMeta, widgets.MediaDefiningClass, ), {})


class PatchRelatedFieldWidgetWrapper(RelatedFieldWidgetWrapper, metaclass=WidgetMonkeyPatchMeta):
    def get_context(self, name, value, attrs, renderer=None):
        # Реализация метода взята из Django 3.2.3
        # Добавлен аргумент renderer и проброшен во внутренний виджет
        from django.contrib.admin.views.main import IS_POPUP_VAR, TO_FIELD_VAR
        rel_opts = self.rel.model._meta
        info = (rel_opts.app_label, rel_opts.model_name)
        self.widget.choices = self.choices
        url_params = '&'.join("%s=%s" % param for param in [
            (TO_FIELD_VAR, self.rel.get_related_field().name),
            (IS_POPUP_VAR, 1),
        ])
        context = {
            'rendered_widget': self.widget.render(name, value, attrs, renderer),
            'is_hidden': self.is_hidden,
            'name': name,
            'url_params': url_params,
            'model': rel_opts.verbose_name,
            'can_add_related': self.can_add_related,
            'can_change_related': self.can_change_related,
            'can_delete_related': self.can_delete_related,
            'can_view_related': self.can_view_related,
        }
        if self.can_add_related:
            context['add_related_url'] = self.get_related_url(info, 'add')
        if self.can_delete_related:
            context['delete_related_template_url'] = self.get_related_url(info, 'delete', '__fk__')
        if self.can_view_related or self.can_change_related:
            context['change_related_template_url'] = self.get_related_url(info, 'change', '__fk__')
        return context

    def render(self, name, value, attrs=None, renderer=None):
        context = self.get_context(name, value, attrs, renderer)
        return self._render(self.template_name, context, renderer)



from django.contrib.admin.views.main import ALL_VAR, ChangeList

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original


class PatchChangeList(ChangeList, metaclass=MonkeyPatchMeta):
    def __init__(self, request, *args, **kwargs):
        self.request = request
        get_original(ChangeList)(self, request, *args, **kwargs)

    @property
    def has_actions(self):
        return bool(self.model_admin.get_actions(self.request))

    @property
    def pagination_required(self):
        return (not self.show_all or not self.can_show_all) and self.multi_page

    @property
    def show_all_url(self):
        need_show_all_link = self.can_show_all and not self.show_all and self.multi_page
        return need_show_all_link and self.get_query_string({ALL_VAR: ""})



# Патч `renderer` должен вызываться после патча `widgets`, т.к. он ссылается
# на файл admin.py приложения auth. Вызовы `admin.register()` должны выполняться
# с уже пропатченным `formfield_overrides`.
from . import (
    changelist,
    filters,
    helpers,
    options,
    prepopulate,
    renderer,
    sites,
    sortable,
    styles,
    tabs,
    widgets,
)



from django import forms
from django.contrib.admin.forms import AdminPasswordChangeForm
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import AdminPasswordChangeForm as AuthPasswordChangeForm
from django.contrib.auth.forms import UserChangeForm, UserCreationForm
from django.forms.forms import DeclarativeFieldsMetaclass

from paper_admin.admin.renderers import PaperFormRenderer
from paper_admin.monkey_patch import MonkeyPatchMeta

# Метакласс MonkeyPatch для класса Form.
FormMonkeyPatchMeta = type("FormMonkeyPatchMeta", (MonkeyPatchMeta, DeclarativeFieldsMetaclass, ), {})

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchAdminPasswordChangeForm(AdminPasswordChangeForm, metaclass=FormMonkeyPatchMeta):
    default_renderer = PaperFormRenderer


class PatchAuthPasswordChangeForm(AuthPasswordChangeForm, metaclass=FormMonkeyPatchMeta):
    default_renderer = PaperFormRenderer


class PatchUserAdmin(UserAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    form = type("UserChangeForm", (UserChangeForm,), {
        "default_renderer": PaperFormRenderer
    })
    add_form = type("UserCreationForm", (UserCreationForm,), {
        "default_renderer": PaperFormRenderer
    })



from django.template import Library, loader

from paper_admin.templatetags.paper_list import (
    ResultRow,
    ResultTable,
    result_list_context,
)

register = Library()


class TreeQueriesResultRow(ResultRow):
    def build_attrs(self):
        attrs = super().build_attrs()
        if self.cl.sortable_allowed:
            attrs["data-parent"] = getattr(self.obj, "parent_id") or 0
        return attrs


class TreeQueriesResultTable(ResultTable):
    row_class = TreeQueriesResultRow


@register.simple_tag(takes_context=True)
def paper_tree_queries_list(context, cl):
    request = context.get("request")
    if request is None:
        return ""

    table = TreeQueriesResultTable(request, cl)
    context = result_list_context(table)
    return loader.render_to_string("admin/tree_queries_change_list_results.html", context, request=request)



"""
Добавляет функционал сортировки методом Drag-n-Drop к ModelAdmin и его дочерним классам.
Реализация опирается на переопределенный класс `ResultList`.

Для того, чтобы сделать модель сортируемой, необходимо добавить в модель числовое поле,
а затем указать название добавленного поля в параметре `sortable` в ModelAdmin:

    # models.py

    class MyModel(models.Model):
        # ...
        order = models.IntegerField("order", default=0)


    # admin.py

    @admin.register(MyModel)
    class MyModelAdmin(admin.ModelAdmin):
        # ...
        sortable = "order"

"""
from . import changelist, helpers, options



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext_lazy as _
from logentry_admin.admin import LogEntryAdmin

from paper_admin.admin.filters import SimpleListFilter
from paper_admin.monkey_patch import MonkeyPatchMeta

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class ContentTypeListFilter(SimpleListFilter):
    title = _('content type')
    parameter_name = 'content_type'
    template = "paper_admin/filters/select.html"

    def lookups(self, request, model_admin):
        return (
            (ct.id, "{}.{}".format(ct.app_label, ct.model))
            for ct in ContentType.objects.order_by("app_label", "model")
        )

    def queryset(self, request, queryset):
        if self.value():
            queryset = queryset.filter(content_type__in=self.value())
        return queryset


class PatchLogEntryAdmin(LogEntryAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    object_history = False
    list_filter = [
        'user',
        ContentTypeListFilter,
        'action_flag'
    ]



from django.contrib.admin.sites import AdminSite
from django.views.i18n import JavaScriptCatalog

from paper_admin.conf import LOCALE_PACKAGES
from paper_admin.monkey_patch import MonkeyPatchMeta, get_original


class PatchAdminSite(AdminSite, metaclass=MonkeyPatchMeta):
    def get_urls(self):
        from django.urls import path
        urlpatterns = get_original(AdminSite)(self)

        # Снимаем требование прав доступа на JS-каталог
        for index, url in enumerate(urlpatterns):
            if url.name == "jsi18n":
                urlpatterns.pop(index)
                urlpatterns.insert(index, path("jsi18n/", self.i18n_javascript, name="jsi18n"))
                break

        return urlpatterns

    def password_change(self, request, extra_context=None):
        # Добавляем шаблонную переменную opts, как в changeform
        from django.contrib.auth import get_user_model
        UserModel = get_user_model()  # noqa: N806
        extra_context = extra_context or {}
        extra_context.setdefault("opts", UserModel._meta)
        return get_original(AdminSite)(self, request, extra_context)

    def i18n_javascript(self, request, extra_context=None):
        return JavaScriptCatalog.as_view(packages=LOCALE_PACKAGES)(request)



import operator
from itertools import zip_longest

from django.contrib.admin import filters
from django.contrib.admin.options import IncorrectLookupParameters
from django.contrib.admin.utils import get_model_from_relation, reverse_field_path
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.dateparse import parse_date
from django.utils.translation import gettext_lazy as _


class SimpleListFilter(filters.SimpleListFilter):
    """
    Отличия от стандартного `SimpleListFilter`:
    1) Используется `request.GET.getlist()` вместо `request.GET.get()`, что позволяет
       указывать несколько значений.
    2) Более универсальный формат `choice`-объекта.
    """
    def __init__(self, request, params, model, model_admin):
        has_value = self.parameter_name in params
        if has_value:
            params.pop(self.parameter_name, None)

        super().__init__(request, params, model, model_admin)

        if has_value:
            values_list = request.GET.getlist(self.parameter_name)
            self.used_parameters[self.parameter_name] = list(filter(lambda x: x != "", values_list))

    def value(self):
        return self.used_parameters.get(self.parameter_name, [])

    def choices(self, changelist):
        yield {
            "selected": not self.value(),
            "value": "",
            "display": _("All"),
        }
        for lookup, title in self.lookup_choices:
            yield {
                "selected": str(lookup) in self.value(),
                "value": lookup,
                "display": title,
            }


class FieldListFilter(filters.ListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field = field
        self.field_path = field_path
        self.title = getattr(field, "verbose_name", field_path)
        super().__init__(request, params, model, model_admin)
        params.pop(self.parameter_name, None)
        values_list = request.GET.getlist(self.parameter_name)
        self.value = list(filter(lambda x: x != "", values_list))

    @property
    def parameter_name(self):
        return self.field_path

    def has_output(self):
        return True

    def get_query(self, value):
        if not value:
            return

        value = [
            None if item == "None" else item
            for item in value
        ]

        if len(value) == 1:
            return models.Q((self.field_path, value[0]))
        else:
            return models.Q(("%s__in" % self.field_path, value))

    def queryset(self, request, queryset):
        query = self.get_query(self.value)
        if query is None:
            return queryset

        try:
            return queryset.filter(query)
        except (ValueError, ValidationError) as e:
            # Fields may raise a ValueError or ValidationError when converting
            # the parameters to the correct type.
            raise IncorrectLookupParameters(e)


class BooleanFieldListFilter(FieldListFilter):
    template = "paper_admin/filters/radio.html"

    def choices(self, changelist):
        yield {
            "selected": not self.value,
            "value": "",
            "display": _("All")
        }
        for lookup, title in (
                ("1", _("Yes")),
                ("0", _("No"))):
            yield {
                "selected": lookup in self.value,
                "value": lookup,
                "display": title,
            }


class ChoicesFieldListFilter(FieldListFilter):
    template = "paper_admin/filters/checkbox.html"

    def choices(self, changelist):
        for lookup, title in self.field.flatchoices:
            yield {
                "selected": str(lookup) in self.value,
                "value": str(lookup),
                "display": title,
            }
        if self.field.null:
            yield {
                "selected": "None" in self.value,
                "value": "None",
                "display": _("None"),
            }


class DateFieldListFilter(FieldListFilter):
    template = "paper_admin/filters/daterange.html"

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field = field
        self.field_path = field_path
        self.title = getattr(field, "verbose_name", field_path)
        super(FieldListFilter, self).__init__(request, params, model, model_admin)

        params.pop(self.start_parameter_name, None)
        params.pop(self.end_parameter_name, None)

        start_values_list = request.GET.getlist(self.start_parameter_name)
        start_values_list = list(filter(lambda x: x != "", start_values_list))

        end_values_list = request.GET.getlist(self.end_parameter_name)
        end_values_list = list(filter(lambda x: x != "", end_values_list))

        self.value = tuple(zip_longest(start_values_list, end_values_list))

    @property
    def start_parameter_name(self):
        return "{}-start".format(self.parameter_name)

    @property
    def end_parameter_name(self):
        return "{}-end".format(self.parameter_name)

    def choices(self, changelist):
        return []

    def get_query(self, value):
        if not value:
            return
        elif len(value) == 1:
            start, end = value[0]

            lookup_since = "%s__gte" % self.field_path
            lookup_until = "%s__lt" % self.field_path

            query = models.Q()
            if start:
                query &= models.Q((lookup_since, parse_date(start)))
            if end:
                query &= models.Q((lookup_until, parse_date(end)))

            return query

        raise ValueError(value)


class RelatedFieldListFilter(FieldListFilter):
    template = "paper_admin/filters/select.html"

    def __init__(self, field, request, params, model, model_admin, field_path):
        other_model = get_model_from_relation(field)
        super().__init__(field, request, params, model, model_admin, field_path)
        self.lookup_choices = self.field_choices(field, request, model_admin)
        if hasattr(field, "verbose_name"):
            self.lookup_title = field.verbose_name
        else:
            self.lookup_title = other_model._meta.verbose_name
        self.title = self.lookup_title

    @property
    def include_empty_choice(self):
        """
        Return True if a "(None)" choice should be included, which filters
        out everything except empty relationships.
        """
        return self.field.null or (self.field.is_relation and self.field.many_to_many)

    def has_output(self):
        if self.include_empty_choice:
            extra = 1
        else:
            extra = 0
        return len(self.lookup_choices) + extra > 1

    def field_choices(self, field, request, model_admin):
        rel_model = field.remote_field.model
        choice_func = operator.attrgetter(
            field.remote_field.get_related_field().attname
            if hasattr(field.remote_field, "get_related_field")
            else "pk"
        )

        queryset = rel_model._default_manager.all()
        is_ordered = bool(queryset.query.order_by or rel_model._meta.ordering)
        if not is_ordered:
            related_admin = model_admin.admin_site._registry.get(
                field.remote_field.model
            )
            if related_admin is not None:
                admin_ordering = related_admin.get_ordering(request)
                queryset = queryset.order_by(*admin_ordering)
        return [(choice_func(x), str(x)) for x in queryset]

    def choices(self, changelist):
        yield {
            "selected": not self.value,
            "value": "",
            "display": _("All")
        }
        for pk, title in self.lookup_choices:
            yield {
                "selected": str(pk) in self.value,
                "value": str(pk),
                "display": title,
            }
        if self.include_empty_choice:
            yield {
                "selected": "None" in self.value,
                "value": "None",
                "display": _("None"),
            }


class AllValuesFieldListFilter(FieldListFilter):
    template = "paper_admin/filters/select.html"

    def __init__(self, field, request, params, model, model_admin, field_path):
        super().__init__(field, request, params, model, model_admin, field_path)
        parent_model, reverse_path = reverse_field_path(model, field_path)
        if model == parent_model:
            queryset = model_admin.get_queryset(request)
        else:
            queryset = parent_model._default_manager.all()
        self.lookup_choices = (
            queryset.distinct().order_by(field.name).values_list(field.name, flat=True)
        )

    def choices(self, changelist):
        yield {
            "selected": not self.value,
            "value": "",
            "display": _("All")
        }
        include_none = False
        for val in self.lookup_choices:
            if val is None:
                include_none = True
                continue
            val = str(val)
            yield {
                "selected": val in self.value,
                "value": val,
                "display": val,
            }
        if include_none:
            yield {
                "selected": "None" in self.value,
                "value": "None",
                "display": _("None"),
            }



from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _


class TreeNodeModelAdminMixin:
    """
    Некоторые аспекты взяты из MPTTModelAdmin.
    """
    change_list_template = "admin/tree_queries_change_list.html"
    list_per_page = 2000  # This will take a really long time to load.
    list_display = ["indented_title"]  # Sane defaults.
    list_display_links = ["indented_title"]  # Sane defaults.
    tree_queries_level_indent = 12

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.with_tree_fields()

    def indented_title(self, obj):
        return format_html(
            '<div style="text-indent:{}px">{}</div>',
            obj.tree_depth * self.tree_queries_level_indent,
            obj
        )

    indented_title.short_description = _("title")


class TreeNodeModelAdmin(TreeNodeModelAdminMixin, admin.ModelAdmin):
    pass



from django import forms
from django.urls import reverse
from mptt.admin import MPTTModelAdmin

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchMPTTModelAdmin(MPTTModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    """
    Фиксы для MPTTModelAdmin.

    https://github.com/darklow/django-suit/issues/381
    """
    list_per_page = 2000
    mptt_level_indent = 12

    def get_changelist(self, request, **kwargs):
        base_changelist_class = get_original(MPTTModelAdmin)(self, request, **kwargs)

        class ChangeList(base_changelist_class):
            def get_ordering(self, request, queryset):
                ordering = super().get_ordering(request, queryset) or []

                if self.sortable_allowed:
                    mptt_opts = self.model._mptt_meta
                    ordering = [mptt_opts.tree_id_attr, mptt_opts.left_attr] + ordering

                return ordering

        return ChangeList

    def _set_order(self, order_dict):
        get_original(MPTTModelAdmin)(self, order_dict)
        self.model._default_manager.rebuild()

    def is_bulk_edit(self, request):
        changelist_url = "admin:%(app_label)s_%(model_name)s_changelist" % {
            "app_label": self.model._meta.app_label,
            "model_name": self.model._meta.model_name,
        }
        return (
            request.path == reverse(changelist_url)
            and request.method == "POST"
            and "_save" in request.POST
        )

    def save_model(self, request, obj, form, change):
        get_original(MPTTModelAdmin)(self, request, obj, form, change)
        if not self.is_bulk_edit(request):
            self.model._default_manager.rebuild()

    def changelist_view(self, request, extra_context=None):
        response = get_original(MPTTModelAdmin)(self, request, extra_context)
        if self.is_bulk_edit(request):
            self.model._default_manager.rebuild()
        return response



import django
from django.contrib.admin.views.main import PAGE_VAR
from django.template import library, loader

register = library.Library()

DOT = '.'
ON_EACH_SIDE = 1
ON_ENDS = 3


@register.simple_tag()
def paper_pagination(cl):
    if not cl.pagination_required:
        return ""

    paginator, page_num = cl.paginator, cl.page_num

    # Use 1-indexed page numbers
    if django.VERSION < (3, 2):
        page_num += 1

    # If there are 5 or fewer pages, display links to every page.
    # Otherwise, do some fancy
    if paginator.num_pages < 5:
        page_range = list(range(1, paginator.num_pages + 1))
    else:
        page_range = []

        if ON_ENDS <= page_num <= paginator.num_pages - ON_ENDS + 1:
            page_range.append(DOT)
            page_range.extend(range(page_num - ON_EACH_SIDE, page_num + ON_EACH_SIDE + 1))
            page_range.append(DOT)
        elif page_num < ON_ENDS:
            page_range.extend(range(1, ON_ENDS + 1))
            page_range.append(DOT)
        elif page_num > paginator.num_pages - ON_ENDS + 1:
            page_range.append(DOT)
            page_range.extend(range(paginator.num_pages - ON_ENDS + 1, paginator.num_pages + 1))

    return loader.render_to_string("admin/pagination.html", {
        "cl": cl,
        "page_num": page_num,
        "page_range": page_range,
    })


@register.simple_tag
def page_url(cl, page_num):
    # Use 1-indexed page numbers
    if django.VERSION < (3, 2):
        page_num -= 1

    return cl.get_query_string({PAGE_VAR: page_num})



from django import forms
from django.contrib.admin.helpers import (
    ActionForm,
    AdminReadonlyField,
    Fieldset,
    InlineAdminForm,
    InlineAdminFormSet,
)
from django.contrib.admin.utils import flatten_fieldsets
from django.forms.forms import DeclarativeFieldsMetaclass
from django.utils.functional import cached_property

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

# Метакласс MonkeyPatch для класса Form.
FormMonkeyPatchMeta = type("FormMonkeyPatchMeta", (MonkeyPatchMeta, DeclarativeFieldsMetaclass, ), {})


class PatchActionForm(ActionForm, metaclass=FormMonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        get_original(ActionForm)(self, *args, **kwargs)
        self.fields["action"].widget.attrs["form"] = "changelist-form"
        self.fields["action"].widget.attrs["class"] = "custom-select"
        self.fields["select_across"].widget.attrs["form"] = "changelist-form"


class PatchAdminReadonlyField(AdminReadonlyField, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        get_original(AdminReadonlyField)(self, *args, **kwargs)
        self.field["contents"] = self.contents()


class PatchFieldset(Fieldset, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        get_original(Fieldset)(self, *args, **kwargs)
        self.has_visible_field = not all(
            field in self.form.fields and self.form.fields[field].widget.is_hidden
            for line in self
            for field in line.fields
        )

    @property
    def media(self):
        return forms.Media()

    @cached_property
    def errors(self):
        # Используется для определения и выделения вкладки, на которой есть ошибки
        errors = []
        for line in self:
            for field in line.fields:
                if field in line.readonly_fields:
                    continue

                for error in self.form[field].errors:
                    errors.append(error)

        return errors


class PatchInlineAdminFormSet(InlineAdminFormSet, metaclass=MonkeyPatchMeta):
    # Шаблон формы перенесен из итератора в отдельное поле empty_form.
    def __iter__(self):
        if self.has_change_permission:
            readonly_fields_for_editing = self.readonly_fields
        else:
            readonly_fields_for_editing = self.readonly_fields + flatten_fieldsets(self.fieldsets)

        # При отправке формсета порядок форм мог быть изменён. Формы, связанные
        # с экземплярами в БД (имеющие `original`) могли оказаться ниже,
        # чем дополнительные (extra_forms) формы. Поэтому для случая связанного
        # формсета мы сопоставляем формы с экземплярами на основании значений
        # поля pk.
        if self.formset.is_bound:
            pk_name = self.formset._pk_field.name

            originals = {
                str(item.pk): item
                for item in self.formset.get_queryset()
            }

            for form in self.formset:
                pk_value = form[pk_name].value()
                if pk_value in originals:
                    original = originals[pk_value]
                    view_on_site_url = self.opts.get_view_on_site_url(original)
                    yield InlineAdminForm(
                        self.formset, form, self.fieldsets, self.prepopulated_fields,
                        original, readonly_fields_for_editing, model_admin=self.opts,
                        view_on_site_url=view_on_site_url,
                    )
                else:
                    yield InlineAdminForm(
                        self.formset, form, self.fieldsets, self.prepopulated_fields,
                        None, self.readonly_fields, model_admin=self.opts,
                    )
        else:
            for form, original in zip(self.formset.initial_forms, self.formset.get_queryset()):
                view_on_site_url = self.opts.get_view_on_site_url(original)
                yield InlineAdminForm(
                    self.formset, form, self.fieldsets, self.prepopulated_fields,
                    original, readonly_fields_for_editing, model_admin=self.opts,
                    view_on_site_url=view_on_site_url,
                )
            for form in self.formset.extra_forms:
                yield InlineAdminForm(
                    self.formset, form, self.fieldsets, self.prepopulated_fields,
                    None, self.readonly_fields, model_admin=self.opts,
                )

    @property
    def empty_form(self):
        return InlineAdminForm(
            self.formset,
            self.formset.empty_form,
            self.fieldsets,
            self.prepopulated_fields,
            None,
            self.readonly_fields,
            model_admin=self.opts,
        )

    @property
    def non_field_errors(self):
        return [
            error
            for form in self.formset.forms
            for error in form.non_field_errors()
        ]



from django import forms
from django.contrib.admin.options import ModelAdmin
from django.utils.functional import cached_property

from paper_admin import conf
from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class AdminTab:
    def __init__(self, request, name, title):
        self.request = request
        self._name = name
        self._title = title
        self.active = False
        self.fieldsets = []
        self.inline_formsets = []

    @property
    def name(self):
        return self._name

    @property
    def title(self):
        return self._title

    @cached_property
    def invalid(self):
        return any(f.errors for f in self.fieldsets) or any(
            f.errors
            for fs in self.inline_formsets
            for f in fs
        )


class PatchModelAdmin(ModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    tabs = [
        (conf.DEFAULT_TAB_NAME, conf.DEFAULT_TAB_TITLE)
    ]

    def get_tabs(self, request, obj=None):
        return list(self.tabs)

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        tabs_config = self.get_tabs(request, obj)
        if conf.DEFAULT_TAB_NAME in dict(tabs_config):
            default_tab_name = conf.DEFAULT_TAB_NAME
        elif tabs_config:
            default_tab_name = tabs_config[0][0]
        else:
            raise RuntimeError("at least one tab required")

        tabs = []
        adminform = context.get("adminform")
        inline_formsets = context.get("inline_admin_formsets")
        for tab_name, tab_title in self.get_tabs(request, obj):
            tab_obj = AdminTab(request, tab_name, tab_title)
            for fieldset in adminform:
                if fieldset.tab == tab_name or (fieldset.tab is None and tab_name == default_tab_name):
                    tab_obj.fieldsets.append(fieldset)
            for inline_formset in inline_formsets:
                if inline_formset.tab == tab_name or (inline_formset.tab is None and tab_name == default_tab_name):
                    tab_obj.inline_formsets.append(inline_formset)
            tabs.append(tab_obj)

        if len(tabs) > 1:
            for tab in tabs:
                if tab.invalid:
                    tab.active = True
                    break
            else:
                tabs[0].active = True
            context["tabs"] = tabs

        return get_original(ModelAdmin)(self, request, context, add=add, change=change, form_url=form_url, obj=obj)



from django.template import Library, loader
from mptt.templatetags.mptt_admin import mptt_items_for_result

from paper_admin.templatetags.paper_list import (
    ResultRow,
    ResultTable,
    result_list_context,
)

register = Library()


class MPTTResultRow(ResultRow):
    def __iter__(self):
        yield from mptt_items_for_result(self.cl, self.obj, self.form)

    def build_attrs(self):
        attrs = super().build_attrs()
        if self.cl.sortable_allowed:
            attrs["data-parent"] = getattr(self.obj, self.obj._mptt_meta.parent_attr + "_id") or 0
        return attrs


class MPTTResultTable(ResultTable):
    row_class = MPTTResultRow


@register.simple_tag(takes_context=True)
def paper_mptt_result_list(context, cl):
    request = context.get("request")
    if request is None:
        return ""

    table = MPTTResultTable(request, cl)
    context = result_list_context(table)
    return loader.render_to_string("admin/mptt_change_list_results.html", context, request=request)



import json

from django import forms
from django.contrib.admin.options import BaseModelAdmin, InlineModelAdmin, ModelAdmin
from django.db import models, router, transaction
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotAllowed,
    JsonResponse,
)
from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchBaseModelAdmin(BaseModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    sortable = None


class PatchInlineModelAdmin(InlineModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if self.sortable and db_field.name == self.sortable:
            kwargs["widget"] = forms.HiddenInput(attrs={
                "class": "paper-formset__order",
            })
        return get_original(InlineModelAdmin)(self, db_field, request, **kwargs)

    def get_ordering(self, request):
        if self.sortable:
            return [self.sortable] + list(get_original(InlineModelAdmin)(self, request))
        return get_original(InlineModelAdmin)(self, request)


class PatchModelAdmin(ModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def __init__(self, model, admin_site):
        get_original(ModelAdmin)(self, model, admin_site)

        if self.sortable:
            # remove sortable field from `list_editable` list
            self.list_editable = self.list_editable or []  # type: list
            if self.sortable in self.list_editable:
                self.list_editable = [
                    field
                    for field in self.list_editable
                    if field != self.sortable
                ]

            # add sortable field to `exclude` list
            self.exclude = self.exclude or []  # type: list
            if self.sortable not in self.exclude:
                self.exclude = list(self.exclude) + [self.sortable]

    def get_sortable_by(self, request):
        if self.sortable and self.has_change_permission(request):
            return []
        else:
            return get_original(ModelAdmin)(self, request)

    def get_urls(self):
        urlpatterns = []
        if self.sortable:
            info = self.model._meta.app_label, self.model._meta.model_name
            urlpatterns.extend([
                path(
                    "set-order/",
                    self.admin_site.admin_view(self.set_order),
                    name="%s_%s_order" % info,
                )
            ])
        return urlpatterns + get_original(ModelAdmin)(self)

    def save_model(self, request, obj, form, change):
        if self.sortable and not change:
            order_value = self.model._default_manager.aggregate(
                max_order=models.Max(self.sortable)
            )["max_order"] or 0
            setattr(obj, self.sortable, order_value + 1)
        get_original(ModelAdmin)(self, request, obj, form, change)

    @csrf_exempt
    def set_order(self, request):
        if request.method != "POST":
            return HttpResponseNotAllowed("Must be a POST request")

        if not self.sortable:
            return HttpResponseForbidden("No sortable field")

        if not self.has_change_permission(request):
            return HttpResponseForbidden("Missing permissions to perform this request")

        try:
            body = json.loads(request.body.decode())
        except json.JSONDecodeError:
            return HttpResponseBadRequest("Invalid JSON")

        with transaction.atomic(using=router.db_for_write(self.model)):
            self._set_order(body)

        return JsonResponse({})

    def _set_order(self, order_dict):
        for pk, order in order_dict.items():
            self.model._default_manager.filter(pk=pk).update(**{
                self.sortable: order
            })



"""
INSTALLED_APPS = [
    "paper_admin",
    "paper_admin.patches.logentry_admin",
    ...
    "logentry_admin",
]

PAPER_MENU = [
    ...
    dict(
        label=_("Logs"),
        icon="fa fa-fw fa-lg fa-history",
        perms="admin.view_logentry",
        models=[
            "admin.LogEntry"
        ]
    ),
]
"""
default_app_config = "paper_admin.patches.logentry_admin.apps.Config"



from django.contrib.admin.views.main import ALL_VAR, ORDER_VAR, SEARCH_VAR
from django.template import Library
from django.template.loader import get_template

FILTER_KEEP_PARAMS = {ALL_VAR, ORDER_VAR, SEARCH_VAR}
register = Library()


@register.simple_tag
def paper_list_filter(cl, spec):
    tpl = get_template(spec.template)
    return tpl.render({
        "title": spec.title,
        "choices": list(spec.choices(cl)),
        "spec": spec,
    })


@register.simple_tag(takes_context=True)
def get_filter_keep_params(context):
    request = context.get("request")
    if request is None:
        return ""

    params = {}
    for key in FILTER_KEEP_PARAMS:
        if key in request.GET:
            params[key] = request.GET.get(key)
    return params.items()



import inspect
import sys

# Pattern for the ids of the original attributes stored.
ORIGINAL_ID = "_monkey_{}__{}"


def get_original(cls=None, attr=None):
    """
    Возвращает оригинальную реализацию атрибута.
    Подобно вызову "super()", можно не указывать ни класс, ни имя атрибута.

        class PatchClass(Class, metaclass=MonkeyPatchMeta):
            def __init__(self, *args, **kwargs):
                get_original()(*args, **kwargs)
    """
    current_frame = inspect.currentframe()
    caller_frame = inspect.getouterframes(current_frame, 2)
    caller_info = caller_frame[1]

    if isinstance(cls, str):
        attr = cls
        cls = None

    if attr is None:
        attr = caller_info.function

    if cls is None:
        cls = type(caller_info.frame.f_locals["self"])

    original_id = ORIGINAL_ID.format(cls.__name__, attr)
    original_methods = getattr(cls, original_id)

    caller_filename = caller_info.frame.f_code.co_filename
    caller_lineno = caller_info.frame.f_code.co_firstlineno
    for index, method in enumerate(original_methods):
        method_filename = method.__code__.co_filename
        method_lineno = method.__code__.co_firstlineno
        if method_filename == caller_filename and method_lineno == caller_lineno:
            return original_methods[index + 1]
    else:
        return original_methods[0]


class MonkeyPatchMeta(type):
    """
    Метакласс, который вместо создания дочернего класса патчит родительский класс.
    Оригинальный метод/атрибут можно получить через вызов get_original.

        class PatchClass(Class, metaclass=MonkeyPatchMeta):
            def __init__(self, *args, **kwargs):
                get_original()(*args, **kwargs)

    """
    def __new__(mcs, name, bases, attrs):  # noqa: N804
        if len(bases) != 1:
            raise TypeError("Monkey patch can only inherit from one base class")

        target = bases[0]

        # fix python 3.8+
        if "__classcell__" in attrs:
            if sys.version_info >= (3, 8):
                attrs["__classcell__"].cell_contents = target

        for attr_name, attr_value in attrs.items():
            if attr_name in {"__module__", "__qualname__", "__classcell__"}:
                continue

            if hasattr(target, attr_name):
                original_id = ORIGINAL_ID.format(target.__name__, attr_name)
                if not hasattr(target, original_id):
                    setattr(target, original_id, [])

                original_methods = getattr(target, original_id)
                original_methods.insert(0, getattr(target, attr_name))

            setattr(target, attr_name, attr_value)

            # for descriptors like "cached_property"
            if hasattr(attr_value, "__set_name__"):
                attr_value.__set_name__(target, attr_name)

        return target



import itertools
import logging
from urllib import parse

from django.contrib.admin import site
from django.shortcuts import resolve_url
from django.utils.text import capfirst

from . import conf

MENU = None
item_counter = itertools.count()
logger = logging.getLogger("paper_admin.menu")


class MenuItemBase:
    __slots__ = [
        "_uid",
        "_parent",
        "_childs",
    ]

    def __init__(self):
        self._uid = next(item_counter)
        self._parent = None
        self._childs = []

    @property
    def uid(self):
        return self._uid

    @property
    def parent(self):
        return self._parent

    @property
    def childs(self):
        return tuple(self._childs)

    def append(self, *args):
        for item in filter(bool, args):
            if not isinstance(item, MenuItemBase):
                continue

            self._childs.append(item)
            item._parent = self


class MenuDivider(MenuItemBase):
    @property
    def divider(self):
        return True


class MenuItem(MenuItemBase):
    __slots__ = [
        "_app",
        "_label",
        "_url",
        "_icon",
        "_active",
        "_classes",
    ]

    def __init__(self, *, app="", label="", url="", icon="", classes=""):
        super().__init__()
        self._app = app
        self._label = label
        self._url = url
        self._classes = classes
        self._icon = icon
        self._active = False

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.label, self.url)

    @property
    def child_items(self):
        return tuple(
            item for item in self._childs
            if isinstance(item, MenuItem)
        )

    @property
    def label(self):
        return capfirst(self._label)

    @label.setter
    def label(self, value):
        if value:
            self._label = value

    @property
    def url(self):
        if not self._url:
            return ""
        elif self._url.startswith(("#", "?")):
            return self._url
        else:
            return resolve_url(self._url)

    @url.setter
    def url(self, value):
        if value:
            self._url = str(value)

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, value):
        if value:
            self._icon = str(value)

    @property
    def classes(self):
        return self._classes

    @classes.setter
    def classes(self, value):
        if value:
            self._classes = str(value)

    @property
    def is_active(self):
        return self._active

    def clean_childs(self):
        self._childs = []

    def activate(self):
        current = self
        while current:
            current._active = True
            current = current.parent

    @classmethod
    def from_site_model(cls, site_model):
        """
        Создание MenuItem из модели admin.site
        """
        return cls(
            label=site_model["name"],
            url=site_model.get("admin_url") or site_model.get("add_url"),
        )

    @classmethod
    def from_site_app_model(cls, app_dict, app_label, model_name):
        """
        Создание MenuItem из модели admin.site по имени приложения и модели
        """
        app = app_dict.get(app_label.lower())
        if app is None:
            return

        try:
            return next(
                cls.from_site_model(site_model)
                for site_model in app["models"]
                if site_model["object_name"].lower() == model_name.lower()
            )
        except StopIteration:
            pass

    @classmethod
    def from_site_app(cls, app_dict, app_label):
        """
        Создание MenuItem из admin.site по имени приложения
        """
        app = app_dict.get(app_label.lower())
        if app is None:
            return

        item = MenuItem(
            label=app["name"],
            url=app["app_url"],
        )
        for model in app["models"]:
            item.append(MenuItem.from_site_model(model))
        return item


def _has_perms(user, perms):
    """
    Проверка наличия прав на пункт меню.

    :type perms: str | list[str] | function
    :rtype: bool
    """
    if not perms:
        return True

    if callable(perms):
        return perms()

    if isinstance(perms, str):
        perms = (perms,)

    for perm in perms:
        if perm == conf.MENU_PERM_SUPERUSER:
            if not user.is_superuser:
                return False
        elif perm == conf.MENU_PERM_STAFF:
            if not user.is_staff:
                return False
        else:
            if not user.has_perm(perm):
                return False

    return True


def _create_menu_item(request, app_dict, conf_item, parent_app=""):
    """
    Создание пункта меню из параметра конфигурации.

    :type request: django.core.handlers.wsgi.WSGIRequest
    :type app_dict: dict[str, dict]
    :type conf_item: str | dict[str, Any]
    :type parent_app: str
    :rtype: MenuItemBase
    """
    if conf_item == conf.MENU_DIVIDER:
        return MenuDivider()

    if isinstance(conf_item, str):
        conf_item = conf_item.lower()
        if "." in conf_item:
            # absolute path to model
            app_label, model_name = conf_item.rsplit(".", 1)
            item = MenuItem.from_site_app_model(app_dict, app_label, model_name)
        elif parent_app:
            # model
            item = MenuItem.from_site_app_model(app_dict, parent_app, conf_item)
        else:
            # app
            item = MenuItem.from_site_app(app_dict, conf_item)
    elif isinstance(conf_item, dict):
        item = MenuItem()

        # проверка прав
        perms = conf_item.get("perms")
        if not _has_perms(request.user, perms):
            return

        app_name = conf_item.get("app")
        site_app = app_dict.get(app_name) if app_name else None
        if site_app:
            item.label = site_app["name"]

        item.label = conf_item.get("label")
        item.url = conf_item.get("url")
        item.icon = conf_item.get("icon")
        item.classes = conf_item.get("classes")

        models = conf_item.get("models")
        if models:
            for conf_subitem in models:
                subitem = _create_menu_item(
                    request, app_dict, conf_subitem, parent_app=app_name or parent_app
                )
                item.append(subitem)
        elif site_app:
            for site_model in site_app["models"]:
                subitem = MenuItem.from_site_model(site_model)
                item.append(subitem)
    else:
        raise TypeError("unsupported type:\n{}".format(conf_item))

    if not item or not item.label:
        return

    if conf.MENU_HIDE_SINGLE_CHILD:
        if len(item.childs) == 1 and not item.childs[0].childs:
            item.url = item.childs[0].url
            item.clean_childs()

    # Возвращаем только пункты с URL или дочерними пунктами
    if item.url or (isinstance(item, MenuItem) and item.child_items):
        return item


def get_menu(request):
    """
    Построение дерева меню.

    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: list[MenuItemBase]
    """
    app_list = site.get_app_list(request)
    app_dict = {
        app["app_label"]: app
        for app in app_list
    }

    if conf.MENU:
        items = (
            _create_menu_item(request, app_dict, item)
            for item in conf.MENU
        )
    else:
        items = (
            MenuItem.from_site_app(app_dict, app["app_label"])
            for app in app_list
        )

    return list(filter(bool, items))


def _compare_urls(url1, url2):
    """
    Сравнение похожести двух урлов

    :type url1: str
    :type url2: str
    :rtype: float
    """
    parsed1 = parse.urlparse(url1)
    path1 = parsed1.path.strip("/").split("/")
    parsed2 = parse.urlparse(url2)
    path2 = parsed2.path.strip("/").split("/")

    max_part_count = max(len(path1), len(path2))
    same_count = sum(
        1 for _ in itertools.takewhile(
            lambda x: x[0] == x[1],
            zip(path1, path2)
        )
    )
    return same_count / max_part_count


def _compare_menu_urls(request, menu):
    """
    Сравнение текущего урла с урлом каждого пункта меню.

    :type request: django.core.handlers.wsgi.WSGIRequest
    :type menu: list[MenuItemBase] | tuple(MenuItemBase)
    :rtype: list[(float, MenuItemBase)]
    """
    result = []
    for item in menu:
        if not isinstance(item, MenuItem):
            continue

        if item.childs:
            subitems = _compare_menu_urls(request, item.childs)
            result.extend(subitems)

        equality = _compare_urls(request.path, item.url)
        result.append((equality, item))
    return result


def activate_menu(request, menu):
    """
    Поиск и активация самого подходящего пункта меню.

    :type request: django.core.handlers.wsgi.WSGIRequest
    :type menu: list of MenuItemBase
    """
    equality_list = _compare_menu_urls(request, menu)
    max_equality = max(pair[0] for pair in equality_list)
    best_items = tuple(
        item
        for equality, item in equality_list
        if equality == max_equality and isinstance(item, MenuItem)
    )

    if len(best_items) == 1:
        # найдено лучшее совпадение
        best_items[0].activate()
    elif len(best_items) > 1:
        # найдено несколько совпадений
        if max_equality == 1:
            for item in best_items:
                item.activate()
        else:
            parent = best_items[0].parent
            if parent is not None:
                parent.activate()
    else:
        # совпадений не найдено
        logger.warning("Not found suitable menu items")



import json

from django.contrib.admin.templatetags.base import InclusionAdminNode
from django.template import Library

register = Library()


def prepopulated_fields_js(context):
    """
    Create a list of prepopulated_fields that should render Javascript for
    the prepopulated fields for both the admin form and inlines.
    """
    prepopulated_fields = {}

    if "adminform" in context:
        adminform = context["adminform"]
        adminform_prepopulated_fields = {}
        for field_name, dependencies in adminform.prepopulated_fields.items():
            slug_field = adminform.form[field_name].field
            adminform_prepopulated_fields[field_name] = {
                "dependencies": dependencies,
                "maxLength": slug_field.max_length or 50,
                "allowUnicode": getattr(slug_field, "allow_unicode", False),
            }
        prepopulated_fields[""] = adminform_prepopulated_fields

    if "inline_admin_formsets" in context:
        for inline_admin_formset in context["inline_admin_formsets"]:
            inline_form = inline_admin_formset.formset.form
            inline_form_prepopulated_fields = {}
            for field_name, dependencies in inline_admin_formset.prepopulated_fields.items():
                slug_field = inline_form.base_fields[field_name]
                inline_form_prepopulated_fields[field_name] = {
                    "dependencies": dependencies,
                    "maxLength": slug_field.max_length or 50,
                    "allowUnicode": getattr(slug_field, "allow_unicode", False),
                }
            prepopulated_fields[inline_admin_formset.formset.prefix] = inline_form_prepopulated_fields

    context.update(
        {"prepopulated_fields_json": json.dumps(prepopulated_fields)}
    )
    return context


@register.tag(name="prepopulated_fields_js")
def prepopulated_fields_js_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=prepopulated_fields_js,
        template_name="prepopulated_fields_js.html",
    )



from django.template import Library
from django.utils.html import escape

register = Library()


@register.simple_tag(takes_context=True)
def paper_row_classes(context, row, model_admin):
    request = context.get("request")
    if request is None:
        return ""

    classes = []
    if hasattr(model_admin, "get_row_classes"):
        classes.extend(model_admin.get_row_classes(request, row.obj))

    return " ".join(map(escape, filter(bool, classes)))


@register.simple_tag(takes_context=True)
def paper_form_classes(context, inline_form):
    request = context.get("request")
    if request is None:
        return ""

    classes = []
    model_admin = inline_form.model_admin
    if hasattr(model_admin, "get_form_classes"):
        classes.extend(model_admin.get_form_classes(request, inline_form.form.instance))

    return " ".join(map(escape, filter(bool, classes)))



from django.contrib.admin import filters
from django.db import models

from paper_admin.admin.filters import (
    AllValuesFieldListFilter,
    BooleanFieldListFilter,
    ChoicesFieldListFilter,
    DateFieldListFilter,
    RelatedFieldListFilter,
)

filters.FieldListFilter.register(
    lambda f: f.remote_field,
    RelatedFieldListFilter,
    take_priority=True,
)

filters.FieldListFilter.register(
    lambda f: isinstance(f, models.BooleanField),
    BooleanFieldListFilter,
    take_priority=True,
)

filters.FieldListFilter.register(
    lambda f: bool(f.choices),
    ChoicesFieldListFilter,
    take_priority=True,
)

filters.FieldListFilter.register(
    lambda f: isinstance(f, models.DateField),
    DateFieldListFilter,
    take_priority=True,
)

filters.FieldListFilter.register(
    lambda f: True,
    AllValuesFieldListFilter,
    take_priority=True,
)



from django.contrib.admin.helpers import Fieldset, InlineAdminFormSet

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original


class PatchFieldset(Fieldset, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        self.tab = kwargs.pop("tab", None)
        get_original(Fieldset)(self, *args, **kwargs)


class PatchInlineAdminFormSet(InlineAdminFormSet, metaclass=MonkeyPatchMeta):
    @property
    def tab(self):
        return getattr(self.opts, "tab", None)



from dal.widgets import WidgetMixin
from dal_select2.widgets import I18N_PATH, Select2WidgetMixin
from django.conf import settings
from django.forms import widgets
from django.utils.safestring import mark_safe

from paper_admin.monkey_patch import MonkeyPatchMeta

# Метакласс MonkeyPatch для класса Widget.
WidgetMonkeyPatchMeta = type("WidgetMonkeyPatchMeta", (MonkeyPatchMeta, widgets.MediaDefiningClass, ), {})


class PatchWidgetMixin(WidgetMixin, metaclass=MonkeyPatchMeta):
    def render(self, name, value, attrs=None, renderer=None, **kwargs):
        # Добавлен параметр renderer
        widget = super(WidgetMixin, self).render(name, value, attrs, renderer=renderer, **kwargs)
        try:
            field_id = attrs['id']
        except (KeyError, TypeError):
            field_id = name
        conf = self.render_forward_conf(field_id)
        return mark_safe(widget + conf)


class PatchSelect2WidgetMixin(Select2WidgetMixin, metaclass=WidgetMonkeyPatchMeta):
    @property
    def media(self):
        extra = "" if settings.DEBUG else ".min"
        i18n_name = self._get_language_code()
        i18n_file = ("%s%s.js" % (I18N_PATH, i18n_name),) if i18n_name else ()

        return widgets.Media(
            js=(
                "autocomplete_light/autocomplete_light%s.js" % extra,
                "autocomplete_light/select2%s.js" % extra,
            ) + i18n_file,
        )



from platform import python_version

from django import get_version
from django.conf import settings
from django.db import connection
from django.utils.translation import gettext_lazy as _

DJANGO_VERSION = get_version()
PYTHON_VERSION = python_version()
DB_VENDOR = connection.display_name

if connection.vendor == "postgresql":
    DB_VERSION = connection.pg_version
elif connection.vendor == "mysql":
    DB_VERSION = ".".join(connection.mysql_version)
elif connection.vendor == "sqlite":
    import sqlite3
    DB_VERSION = sqlite3.version
else:
    DB_VERSION = ""

FAVICON = getattr(settings, "PAPER_FAVICON", "paper_admin/dist/assets/default_favicon.png")

ENVIRONMENT_NAME = getattr(settings, "PAPER_ENVIRONMENT_NAME", "")
ENVIRONMENT_COLOR = getattr(settings, "PAPER_ENVIRONMENT_COLOR", "")

MENU = getattr(settings, "PAPER_MENU", None)
MENU_DIVIDER = getattr(settings, "PAPER_MENU_DIVIDER", "-")
MENU_PERM_STAFF = getattr(settings, "PAPER_MENU_PERM_STAFF", "staff")
MENU_PERM_SUPERUSER = getattr(settings, "PAPER_MENU_PERM_SUPERUSER", "superuser")
MENU_HIDE_SINGLE_CHILD = getattr(settings, "PAPER_MENU_HIDE_SINGLE_CHILD", True)

DEFAULT_TAB_NAME = getattr(settings, "PAPER_DEFAULT_TAB_NAME", "general")
DEFAULT_TAB_TITLE = getattr(settings, "PAPER_DEFAULT_TAB_TITLE", _("General"))

LOCALE_PACKAGES = getattr(settings, "PAPER_LOCALE_PACKAGES", ["paper_admin", "django.contrib.admin"])



import json

from django import forms
from django.contrib import messages
from django.contrib.admin.options import (
    IS_POPUP_VAR,
    InlineModelAdmin,
    ModelAdmin,
    csrf_protect_m,
)
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.contrib.admin.utils import model_format_dict
from django.contrib.admin.views.main import SEARCH_VAR
from django.contrib.auth import get_permission_codename, get_user_model
from django.db.models.fields import BLANK_CHOICE_DASH
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext as _

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchModelAdmin(ModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    list_per_page = 20
    object_history = True  # show "History" button
    changelist_tools = True  # show buttons in changelist view
    changelist_tools_template = "paper_admin/includes/changelist_tools.html"

    @property
    def media(self):
        return forms.Media()

    def get_action_choices(self, request, default_choices=BLANK_CHOICE_DASH):
        # Изменение placeholder у выпадающего списка
        choices = [("", _("Action"))]
        for func, name, description in self.get_actions(request).values():
            choice = (name, description % model_format_dict(self.opts))
            choices.append(choice)
        return choices

    @csrf_protect_m
    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context["search_var"] = SEARCH_VAR
        return get_original(ModelAdmin)(self, request, extra_context=extra_context)

    def history_view(self, request, object_id, extra_context=None):
        # Добавление в список логов ссылки на пользователя и ссылки на запись лога
        from django.contrib.admin.models import LogEntry

        log_opts = LogEntry._meta
        codename = get_permission_codename("view", log_opts)
        has_log_change_permission = request.user.has_perm(
            "%s.%s" % (log_opts.app_label, codename)
        )
        log_model_registered = self.admin_site.is_registered(LogEntry)

        UserModel = get_user_model()  # noqa: N806
        user_opts = UserModel._meta
        codename = get_permission_codename("view", user_opts)
        has_user_change_permission = request.user.has_perm(
            "%s.%s" % (user_opts.app_label, codename)
        )
        user_model_registered = self.admin_site.is_registered(UserModel)

        default_extra = {
            "log_opts": log_opts,
            "user_opts": user_opts,
            "show_log_link": has_log_change_permission and log_model_registered,
            "show_user_link": has_user_change_permission and user_model_registered
        }
        default_extra.update(extra_context or {})
        return get_original(ModelAdmin)(self, request, object_id, default_extra)

    def response_delete(self, request, obj_display, obj_id):
        # Добавление редиректа на список при наличии прав на чтение
        opts = self.model._meta

        if IS_POPUP_VAR in request.POST:
            popup_response_data = json.dumps({
                "action": "delete",
                "value": str(obj_id),
            })
            return TemplateResponse(request, self.popup_response_template or [
                "admin/%s/%s/popup_response.html" % (opts.app_label, opts.model_name),
                "admin/%s/popup_response.html" % opts.app_label,
                "admin/popup_response.html",
            ], {
                "popup_response_data": popup_response_data,
            })

        self.message_user(
            request,
            _("The %(name)s “%(obj)s” was deleted successfully.") % {
                "name": opts.verbose_name,
                "obj": obj_display,
            },
            messages.SUCCESS,
        )

        if self.has_view_permission(request, None):
            post_url = reverse(
                "admin:%s_%s_changelist" % (opts.app_label, opts.model_name),
                current_app=self.admin_site.name,
            )
            preserved_filters = self.get_preserved_filters(request)
            post_url = add_preserved_filters(
                {"preserved_filters": preserved_filters, "opts": opts}, post_url
            )
        else:
            post_url = reverse("admin:index", current_app=self.admin_site.name)

        return HttpResponseRedirect(post_url)

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        response = get_original(ModelAdmin)(self, request, context, add=add, change=change, form_url=form_url, obj=obj)

        # Перенос кода из `admin_modify.submit_row`.
        _context = response.context_data
        add = _context["add"]
        change = _context["change"]
        is_popup = _context["is_popup"]
        save_as = _context["save_as"]
        show_save = _context.get("show_save", True)
        show_save_and_add_another = _context.get("show_save_and_add_another", True)
        show_save_and_continue = _context.get("show_save_and_continue", True)

        has_add_permission = _context["has_add_permission"]
        has_change_permission = _context["has_change_permission"]
        has_view_permission = _context["has_view_permission"]
        has_editable_inline_admin_formsets = _context["has_editable_inline_admin_formsets"]

        can_save = (
            (has_change_permission and change)
            or (has_add_permission and add)
            or has_editable_inline_admin_formsets
        )

        _context.update({
            "can_change": (
                has_change_permission
                or has_editable_inline_admin_formsets
            ),
            "show_delete_link": (
                not is_popup
                and _context["has_delete_permission"]
                and change
                and _context.get("show_delete", True)
            ),
            "show_save_as_new": (
                not is_popup
                and has_change_permission
                and change
                and save_as
            ),
            "show_save_and_add_another": (
                has_add_permission
                and not is_popup
                and (not save_as or add)
                and can_save
                and show_save_and_add_another
            ),
            "show_save_and_continue": (
                not is_popup
                and can_save
                and has_view_permission
                and show_save_and_continue
            ),
            "show_save": show_save and can_save,
            "show_close": not (show_save and can_save)
        })

        # Показывать ли кнопку логов
        _context["show_history"] = self.object_history

        return response


class PatchInlineModelAdmin(InlineModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    @property
    def media(self):
        return forms.Media()



from collections import OrderedDict

from django.contrib.admin.views.main import ChangeList

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original


class PatchChangeList(ChangeList, metaclass=MonkeyPatchMeta):
    @property
    def sortable_allowed(self):
        return bool(self.model_admin.sortable and self.model_admin.has_change_permission(self.request))

    def get_ordering(self, request, queryset):
        if self.sortable_allowed:
            return [self.model_admin.sortable, "-" + self.model._meta.pk.name]
        else:
            return get_original(ChangeList)(self, request, queryset)

    def get_ordering_field_columns(self):
        if self.sortable_allowed:
            return OrderedDict()
        else:
            return get_original(ChangeList)(self)



import json

import django
from django.contrib.admin.widgets import AutocompleteMixin, ManyToManyRawIdWidget
from django.forms import widgets

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original

# Метакласс MonkeyPatch для класса Widget.
WidgetMonkeyPatchMeta = type("WidgetMonkeyPatchMeta", (MonkeyPatchMeta, widgets.MediaDefiningClass, ), {})


class PatchInput(widgets.Input, metaclass=WidgetMonkeyPatchMeta):
    def __init__(self, attrs=None):
        default_attrs = {
            "class": "form-control form-control-lg"
        }
        if attrs:
            default_attrs.update(attrs)
        get_original(widgets.Input)(self, attrs=default_attrs)


class PatchSelect(widgets.Select, metaclass=WidgetMonkeyPatchMeta):
    def __init__(self, attrs=None, choices=()):
        default_attrs = {
            "class": "custom-select custom-select-lg"
        }
        if attrs:
            default_attrs.update(attrs)
        get_original(widgets.Select)(self, attrs=default_attrs, choices=choices)


class PatchTextarea(widgets.Textarea, metaclass=WidgetMonkeyPatchMeta):
    def __init__(self, attrs=None):
        default_attrs = {
            "class": "form-control form-control-lg"
        }
        if attrs:
            default_attrs.update(attrs)
        get_original(widgets.Textarea)(self, attrs=default_attrs)


class PatchSelectDateWidget(widgets.SelectDateWidget, metaclass=WidgetMonkeyPatchMeta):
    def get_context(self, name, value, attrs):
        # disable select2 allowClear
        attrs = attrs or {}
        attrs.setdefault("data-allow-clear", "false")
        return get_original(widgets.SelectDateWidget)(self, name, value, attrs)


class PatchManyToManyRawIdWidget(ManyToManyRawIdWidget, metaclass=WidgetMonkeyPatchMeta):
    def __init__(self, rel, admin_site, attrs=None, using=None):
        get_original(ManyToManyRawIdWidget)(self, rel, admin_site, attrs=attrs, using=using)
        self.attrs["class"] = (self.attrs.get("class", "") + " vManyToManyRawIdAdminField").strip()

    def get_context(self, name, value, attrs):
        # удаление CSS-класса
        return super().get_context(name, value, attrs)


class PatchAutocompleteMixin(AutocompleteMixin, metaclass=MonkeyPatchMeta):
    def __init__(self, field, admin_site, attrs=None, choices=(), using=None):
        default_attrs = {
            "class": "custom-select custom-select-lg"
        }
        if attrs:
            default_attrs.update(attrs)
        get_original(AutocompleteMixin)(self, field, admin_site, attrs=default_attrs, choices=choices, using=using)

    @property
    def media(self):
        return widgets.Media()

    def build_attrs(self, base_attrs, extra_attrs=None):
        attrs = super().build_attrs(base_attrs, extra_attrs=extra_attrs)
        attrs.setdefault("class", "")
        attrs.update({
            "data-width": "",
            "data-ajax--cache": "true",
            "data-ajax--type": "GET",
            "data-ajax--url": self.get_url(),
            "data-allow-clear": json.dumps(not self.is_required),
            "data-placeholder": "",  # Allows clearing of the input.
            "class": attrs["class"] + (" " if attrs["class"] else "") + "admin-autocomplete",
        })

        if django.VERSION >= (3, 2):
            attrs.update({
                "data-app-label": self.field.model._meta.app_label,
                "data-model-name": self.field.model._meta.model_name,
                "data-field-name": self.field.name,
            })

        # Django 4.0 support
        if hasattr(self, "lang"):
            attrs["lang"] = self.i18n_name

        return attrs



from django import forms
from django.contrib.admin.options import InlineModelAdmin, ModelAdmin

from paper_admin.monkey_patch import MonkeyPatchMeta

# Метакласс MonkeyPatch для класса BaseModelAdmin.
ModelAdminMonkeyPatchMeta = type("ModelAdminMonkeyPatchMeta", (MonkeyPatchMeta, forms.MediaDefiningClass), {})


class PatchModelAdmin(ModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def get_row_classes(self, request, obj):
        return []


class PatchInlineModelAdmin(InlineModelAdmin, metaclass=ModelAdminMonkeyPatchMeta):
    def get_form_classes(self, request, obj):
        return []



from django.contrib.auth.forms import PasswordResetForm, SetPasswordForm

from .renderers import PaperFormRenderer


class AdminPasswordResetForm(PasswordResetForm):
    default_renderer = PaperFormRenderer


class AdminSetPasswordForm(SetPasswordForm):
    default_renderer = PaperFormRenderer



from django.template import library, loader

from ..menu import activate_menu, get_menu

register = library.Library()


@register.simple_tag(takes_context=True)
def paper_menu(context, parent_id):
    request = context.get("request")
    menu = get_menu(request)
    activate_menu(request, menu)
    return loader.render_to_string(
        "paper_admin/_menu.html",
        {
            "level": 1,
            "items": menu,
            "parent": parent_id,
        },
        request=request
    )



from django.contrib.admin.helpers import InlineAdminFormSet

from paper_admin.monkey_patch import MonkeyPatchMeta, get_original


class PatchInlineAdminFormSet(InlineAdminFormSet, metaclass=MonkeyPatchMeta):
    def __init__(self, *args, **kwargs):
        get_original(InlineAdminFormSet)(self, *args, **kwargs)
        self.sortable_allowed = self.opts.sortable and self.has_change_permission



from django.contrib.admin import site
from django.contrib.auth import get_user_model
from django.contrib.auth import views as auth_views

from .forms import AdminPasswordResetForm, AdminSetPasswordForm


class AdminViewMixin:
    def get_context_data(self, **kwargs):
        data = super().get_context_data(**kwargs)
        data.update(site.each_context(self.request))

        UserModel = get_user_model()  # noqa: N806
        data.setdefault("opts", UserModel._meta)

        return data


class AdminPasswordResetView(AdminViewMixin, auth_views.PasswordResetView):
    form_class = AdminPasswordResetForm


class AdminPasswordResetDoneView(AdminViewMixin, auth_views.PasswordResetDoneView):
    pass


class AdminPasswordResetConfirmView(AdminViewMixin, auth_views.PasswordResetConfirmView):
    form_class = AdminSetPasswordForm


class AdminPasswordResetCompleteView(AdminViewMixin, auth_views.PasswordResetCompleteView):
    pass
